/*
** curses.c - Mac curses
** Copyright (c) 1996, by Sheldon Simms
**
** This software is freely distributable and may be copied or modified
** for any purpose. But I maintain copyright on this source code and
** if you modify or redistribute this program you must include the
** statement "Copyright (c) 1996 by Sheldon Simms", if merely redistributed,
** or "portions Copyright (c) 1996 by Sheldon Simms" if modified.
*/

#include <signal.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "xcurses.h"

#define ERR 0
#define OK  1

#define LF 0x0a
#define CR 0x0d
#define INT 0x03

#define M_CBREAK 1	/* In this implementation CBREAK == RAW */
#define M_RAW    2
#define M_DELAY  4

#define W_WIN    1
#define W_SUBWIN 2
#define W_PAD    3

/* Global variables start here */

WINDOW *stdscr;
WINDOW *curscr;
WINDOW *windowHead;

int pLINES = 24, pCOLS = 80;
int LINES, COLS;
int f_echo = TRUE, f_nl = TRUE;
int inputMode = M_CBREAK;
unsigned char strBuf[ 256 ]; /* A temporary buffer for strings. The length is somewhat arbitrary */

int readByte;

/* Used by longname() call. Basically a waste of space */

char *terminalName = "Mac curses window, flexible character set";

/*
** scrollWindow takes advantage of the fact that the window contents are
** stored in a single block of memory < win->lines * win->cols > long
** consisting of < win->lines > lines stored in order through the block.
** Scrolling the window, then consists of shifting the bytes < win->cols >
** locations back towards the beginning of the block, which has the effect
** of scrolling the entire window up one line.
**
** Addition: There is a check for subwindows. Since subwindows use another
** window's buffer they can't be scrolled in that way. They are scrolled
** by looping on y and x and moving each character to y-1,x.
*/

void scrollWindow( WINDOW *win, bool update )
{
	short i, x, y, wl;
	short *wbl, *wel;
	chtype *src, *dst, *ptr_to_last_line;

	wl = win->lines;						/* get some values to minimize pointer derefs */
	ptr_to_last_line = win->winBuf[ wl - 1 ];

	if ( update )
	{
		wbl = win->begLine;
		wel = win->endLine;
		for ( i = 0; i < wl; i++ )
		{
			wbl[ i ] = 0;					/* set update markers to show that every line */
			wel[ i ] = win->cols - 1;		/* potentially needs updating */
		}
	}

	if ( win->type != W_SUBWIN )
	{
		dst = win->winBuf[ 0 ];				/* move window contents data back */
		src = dst + win->cols;				/* win->cols bytes */
		while ( dst < ptr_to_last_line )
			*dst++ = *src++;

		for( i = 0; i < win->cols; i++ )	/* leave blank line at bottom */
			*dst++ = ' ';
			/* *dst++ = win->attrword | ' '; */
	}
	else
	{
		for ( y = 1; y < win->lines; y++ )
			for ( x = 0; x < win->cols; x++ )
				win->winBuf[ y - 1 ][ x ] = win->winBuf[ y ][ x ];

		for ( x = 0; x < win->cols; x++ )
			win->winBuf[ y - 1 ][ x ] = ' ';
			/* win->winBuf[ y - 1 ][ x ] = win->attrword | ' '; */
	}
}

/*
** consolePutchar handles outputting a character and updating cursor
** position, mapping CR->LF etc.
*/

void consolePutchar( WINDOW *win, chtype ch )
{
	short wcx, wcy;

	if ( ch > 0x1f )
	{
		wcx = win->curX;						/* save values to minimize pointer derefs */
		wcy = win->curY;
		
		win->winBuf[ wcy ][ wcx ] = ch |= win->attrword;

		if ( wcx <= win->begLine[ wcy ] )	/* adjust update markers */
			win->begLine[ wcy ] = wcx;
		if ( wcx >= win->endLine[ wcy ] )
			win->endLine[ wcy ] = wcx;
				
		if ( wcx < ( win->cols - 1 ))
			win->curX++;
		else
		{
			win->curX = 0;
			if ( wcy == ( win->lines - 1 ))
			{
				if ( win->f_scroll )
					scrollWindow( win, TRUE );
			}
			else
				win->curY++;
		}
	}
	else if ( ch == LF )
	{
		if ( f_nl )								/* add effect of CR if nl() */
			win->curX = 0;

		if ( win->curY < ( win->lines - 1 ))
			win->curY++;
		else if ( win->f_scroll )
		{
			scrollWindow( win, TRUE );
			win->curX = 0;
		}
	}
	else if ( ch == CR )
		win->curX = 0;
}

/*
** consoleGetchar is a pretty complex function. It gets a character from
** the keyboard using the DRead call. DRead is slow, but how fast can
** people type? Anyway consoleGetchar also provides a blinking cursor
** if leaveok() has not been called, maps CR->LF if nl(), advances the
** cursor position, and echoes the typed character if echo().
*/

int consoleGetchar( WINDOW *win )
{
	chtype screenChar;
	bool toggle, toggle_changed; /* used for blinking cursor */
	short gl_wcx, gl_wcy, wcx, wcy;
	short result = -1;
	clock_t oldtime, newtime;

	gl_wcy = win->top + win->curY;
	gl_wcx = win->left + win->curX;

	/* save character from screen so that cursor can be blinked */
	screenChar = curscr->winBuf[ gl_wcy ][ gl_wcx ];
	toggle = FALSE;
    toggle_changed = TRUE;
	oldtime = clock();

	while( (short)result < (short)0 )
	{
		/* if we're in nodelay mode, there is no cursor since we */
		/* won't be hanging around waiting for a keypress */
		if ( !win->f_leave && !win->f_nodelay )
		{
          if (toggle_changed)
            {
              /* Make a blinking cursor */
              if ( !toggle )
				mac_putchar( '_', gl_wcx, gl_wcy ); 		/* draw underline cursor */
              else
				mac_putchar( screenChar, gl_wcx, gl_wcy );	/* redraw screen character */

              mac_update( gl_wcy, gl_wcx, gl_wcx );

              toggle_changed = FALSE;
            }
		}

		result = mac_getchar();

		/* in nodelay mode we return whether we got a character or not */
		if ( win->f_nodelay )
			break;

		/* see if it's time to blink yet */
		newtime = clock();
		if(( CLOCKS_PER_SEC / 2 ) <= ( newtime - oldtime ))
		{
			toggle = !toggle;
            toggle_changed = TRUE;
			oldtime = newtime;
		}
	}

	if ( result < 0 )
		return -1;	/* can only happen in nodelay mode */

	if ( result == INT )
		raise( SIGINT );
		
	/* map CR->LF if in nl mode */
	if ( result == CR && f_nl )
		result = LF;

	if ( !f_echo || result < 0x20 )
	{
		/* if we're not echoing characters we have to put the old screen */
		/* character back. We also have to do this if the user typed a   */
		/* non-printing character, even if we are in echo mode */
		mac_putchar( screenChar, gl_wcx, gl_wcy );	/* redraw screen character */
		mac_update( gl_wcy, gl_wcx, gl_wcx );
	}
	else if ( result > 0x1f )
	{
		/* echo typed character */
		wcx = win->curX;
		wcy = win->curY;
		
		/* put the character out there now */
		mac_putchar( win->attrword | result, gl_wcx, gl_wcy );
		mac_update( gl_wcy, gl_wcx, gl_wcx );
		win->winBuf[ wcy ][ wcx ] = curscr->winBuf[ wcy ][ wcx ] =
			(chtype)( win->attrword | result );
			
		if ( wcx < ( win->cols - 1 ))
			win->curX++;
	}
	else if ( result == CR )
		win->curX = 0;
	else if ( result == LF )
	{
		if ( win->curY < ( win->lines - 1 ))
			win->curY++;
	}

	return (int)result;
}

/*
** allocWindow is all sorts of fun. We get to allocate a window
** structure, allocate two arrays to keep track of the beginning
** and ending of modified areas on each line of the window, and
** allocate a two-dimensional buffer that holds the character
** data of the window. Then we fill in default values for all the
** various fields
*/

WINDOW *allocWindow( int y, int x, int lines, int cols )
{
	int i;
	WINDOW *newWin;

	/* get the window structure */
	newWin = (WINDOW *)malloc( sizeof( WINDOW ));
	if ( newWin == NULL )
	{
		return (WINDOW *)NULL;
	}

	newWin->begLine = (short *)malloc( lines * sizeof( short ));
	if ( newWin->begLine == NULL )
	{
		free( newWin );
		return (WINDOW *)NULL;
	}

	newWin->endLine = (short *)malloc( lines * sizeof( short ));
	if ( newWin->endLine == NULL )
	{
		free( newWin->begLine );
		free( newWin );
		return (WINDOW *)NULL;
	}

	/* pointers to the start of each line in the winBuf */
	newWin->winBuf = (chtype **)malloc( lines * sizeof( chtype * ));
	if ( newWin->winBuf == NULL )
	{
		free( newWin->endLine );
		free( newWin->begLine );
		free( newWin );
		return (WINDOW *)NULL;
	}

	/* get the winBuf */
	newWin->winBuf[ 0 ] = (chtype *)malloc( lines * cols * sizeof( chtype ));
	if ( newWin->winBuf[ 0 ] == NULL )
	{
		free( newWin->winBuf );
		free( newWin->endLine );
		free( newWin->begLine );
		free( newWin );
		return (WINDOW *)NULL;
	}

	/* set up the buffer so it can be accessed with array notation */
	/* and initialize the update arrays */
	for( i = 1; i < lines; i++ )
	{
		newWin->winBuf[ i ] = newWin->winBuf[ 0 ] + ( i * cols );
		newWin->begLine[ i ] = cols - 1;
		newWin->endLine[ i ] = -1;
	}

	newWin->type = W_WIN;
	newWin->f_scroll = FALSE;
	newWin->f_clear = FALSE;
	newWin->f_leave = FALSE;
	newWin->f_standout = FALSE;
	newWin->f_nodelay = FALSE;
	newWin->curX = 0;
	newWin->curY = 0;
	newWin->lines = lines;
	newWin->cols = cols;
	newWin->left = x;
	newWin->top = y;
	newWin->right = x + cols;
	newWin->bottom = y + lines;
	newWin->attrword = 0x0000;
	newWin->next = windowHead;
	windowHead = newWin;

	return newWin;
}

/* same as allocWindow. instead of allocating a winBuf though, */
/* use orig's winBuf. All we have to do is allocate the array  */
/* of pointers to rows and then point at the appropriate place */
/* in orig's winBuf. */

WINDOW *subwin( WINDOW *orig, int lines, int cols, int y, int x )
{
	int i, sx, sy;
	WINDOW *newWin;

	/* We're given the subwin's location in global coordinates, so */
	/* we have to change them to orig-coordinates */

	sx = x - orig->left;
	sy = y - orig->top;
	if (( sx + cols ) > orig->cols || ( sy + lines ) > orig->lines )
		return NULL;

	newWin = (WINDOW *)malloc( sizeof( WINDOW ));
	if ( newWin == NULL )
	{
		return (WINDOW *)NULL;
	}

	newWin->begLine = (short *)malloc( lines * sizeof( short ));
	if ( newWin->begLine == NULL )
	{
		free( newWin );
		return (WINDOW *)NULL;
	}

	newWin->endLine = (short *)malloc( lines * sizeof( short ));
	if ( newWin->endLine == NULL )
	{
		free( newWin->begLine );
		free( newWin );
		return (WINDOW *)NULL;
	}

	newWin->winBuf = (chtype **)malloc( lines * sizeof( chtype * ));
	if ( newWin->winBuf == NULL )
	{
		free( newWin->endLine );
		free( newWin->begLine );
		free( newWin );
		return (WINDOW *)NULL;
	}

	newWin->winBuf[ 0 ] = orig->winBuf[ sy ] + sx;
	for( i = 1; i < lines; i++ )
	{
		newWin->winBuf[ i ] = newWin->winBuf[ 0 ] + ( i * orig->cols );
		newWin->begLine[ i ] = cols - 1;
		newWin->endLine[ i ] = -1;
	}

	newWin->type = W_SUBWIN;
	newWin->f_scroll = FALSE;
	newWin->f_clear = FALSE;
	newWin->f_leave = FALSE;
	newWin->f_standout = FALSE;
	newWin->f_nodelay = FALSE;
	newWin->curX = 0;
	newWin->curY = 0;
	newWin->lines = lines;
	newWin->cols = cols;
	newWin->left = x;
	newWin->top = y;
	newWin->right = x + cols;
	newWin->bottom = y + lines;
	newWin->attrword = 0x0000;
	newWin->next = windowHead;
	windowHead = newWin;

	return newWin;
}

/* pretty straightforward... */

void freeWindow( WINDOW *win )
{
	if ( win->type != W_SUBWIN )
		free( win->winBuf[ 0 ] );
	free( win->winBuf );
	free( win->endLine );
	free( win->begLine );
	free( win );
}

/* initscr sets up the console driver, allocates curscr and stdscr */
/* and initializes LINES and COLS */

int initscr( void )
{
	int result;
	
	result = init_mac_curses();
	if ( result < 0 )
		return ERR;
		
	LINES = pLINES;
	COLS = pCOLS;
		
	curscr = newwin( 0, 0, 0, 0 );
	if ( curscr == NULL )
		return ERR;

	stdscr = newwin( 0, 0, 0, 0 );
	if ( curscr == NULL )
	{
		freeWindow( curscr );
		return ERR;
	}
	stdscr->f_clear = TRUE;

	return OK;
}

/* endwin restores the console driver to the state we found at */
/* startup and frees any windows left hanging around */

int endwin( void )
{
	WINDOW *p, *b;

	p = windowHead;
	while ( p != NULL )		/* free windows */
	{
		b = p;
		p = p->next;
		freeWindow( b );
	}
	close_mac_curses();

	return OK;
}

/* Get a new window. The work is actually all done in allocWindow */
/* but newwin() has the feature that specifying 0 for the height  */
/* or width of the window gives you LINES or COLS respectively.   */
/* That feature is implemented here. */

WINDOW *newwin( int lines, int cols, int uly, int ulx )
{
	WINDOW *nwin;

	if ( lines > LINES || cols > COLS )
		return (WINDOW *)NULL;

	if ( lines == 0 )
		lines = LINES - uly;
	if ( cols == 0 )
		cols = COLS - ulx;

	nwin = allocWindow( uly, ulx, lines, cols );
	if ( nwin != NULL )
		werase( nwin );

	return nwin;
}

int mvwin( WINDOW *win, int y, int x )
{
	int diffy, diffx;

	if (( y + win->lines ) > LINES || ( x + win->cols ) > COLS )
		return ERR;

	diffy = y - win->top;
	diffx = x - win->left;
	win->left += diffx;
	win->right += diffx;
	win->top += diffy;
	win->bottom += diffy;

	return OK;
}

int delwin( WINDOW *win )
{
	freeWindow( win );
	return OK;
}

/* touchwin sets the update arrays so that the update region of */
/* every line starts at 0 and ends at win->cols-1: the update   */
/* region becomes the entire window. BTW, window coordinates go */
/* from 0 to win->lines - 1 in the vertical direction and from  */
/* 0 to win->cols - 1 in the horizontal direction */

int touchwin( WINDOW *win )
{
	short i, wc;

	wc = win->cols - 1;
	for ( i = 0; i < win->lines; i++ )
	{
		win->begLine[ i ] = 0;
		win->endLine[ i ] = wc;
	}

	return OK;
}

/* touch line does the same as touchwin but only for a specified */
/* number of lines starting at line # start */

int touchline( WINDOW *win, int start, int count )
{
	short i, wc;

	wc = win->cols - 1;
	for ( i = start; i < ( start + count ); i++ )
	{
		win->begLine[ i ] = 0;
		win->endLine[ i ] = wc;
	}

	return OK;
}

/*
My first thought was that overlay() and overwrite() only copied the
part of the source window that overlapped the dest window. I now think
that is incorrect but the code to compute the overlap in global
coordinates is here so I didn't delete it. It's not being used now,
but it's useful code that would be used in some of the enhanced curses
functions I didn't implement.

int overLayWrite( WINDOW *srcwin, WINDOW *dstwin, bool destructive )
{
	int sx, sy, dx, dy;
   int sl, st, sr, sb;
   int dl, dt, dr, db;
   chtype **sbuf, **dbuf, ch;

   sl = st = sr = sb = 0;
   dl = dt = dr = db = 0;
   sbuf = srcwin->winBuf;
   dbuf = dstwin->winBuf;

   if ( dstwin->left >= srcwin->left && dstwin->left <= srcwin->right )
	    sl = dstwin->left;
	else if ( srcwin->left >= dstwin->left && srcwin->left <= dstwin->right )
	    sl = srcwin->left;

	if ( dstwin->right <= srcwin->right && dstwin->right >= srcwin->left )
	    sr = dstwin->right;
   else if ( srcwin->right <= dstwin->right && srcwin->right >= dstwin->left )
	    sr = srcwin->right;

   if ( dstwin->top >= srcwin->top && dstwin->top <= srcwin->bottom )
	    st = dstwin->top;
	else if ( srcwin->top >= dstwin->top && srcwin->top <= dstwin->bottom )
	    st = srcwin->top;

	if ( dstwin->bottom <= srcwin->bottom && dstwin->bottom >= srcwin->top )
	    sb = dstwin->bottom;
   else if ( srcwin->bottom <= dstwin->bottom && srcwin->bottom >= dstwin->top )
	    sb = srcwin->bottom;

	if ( !sl && !st && !sr && !sb )
	    return OK;

	dl = sl - dstwin->left;
   dt = st - dstwin->top;
   dr = sr - dstwin->right;
   db = sb - dstwin->bottom;

	sl = sl - srcwin->left;
   st = st - srcwin->top;
   sr = sr - srcwin->right;
   sb = sb - srcwin->bottom;

   for ( sy = st, dy = dt; sy < sb && dy < db; sy++, dy++ )
   	for ( sx = sl, dx = dl; sx < sr && dx < dr; sx++, dx++ )
	    {
	        ch = sbuf[ sy ][ sx ];
	        if ( destructive || ( ch & 0xff ) > ' ' )
	        	dbuf[ dy ][ dx ] = ch;
		}
}
*/

/* copy data from source window to dest window except where source */
/* window has a space */

int overlay( WINDOW *srcwin, WINDOW *dstwin )
{
	chtype ch;
	int y, x;

	for ( y = 0; y < srcwin->lines && y < dstwin->lines; y++ )
		for ( x = 0; x < srcwin->cols && x < dstwin->cols; x++ )
		{
			ch = srcwin->winBuf[ y ][ x ];
			if(( ch & 0xff ) > ' ' )
				dstwin->winBuf[ y ][ x ] = ch;
		}
	return OK;
}

/* copy source to dest, including spaces */

int overwrite( WINDOW *srcwin, WINDOW *dstwin )
{
	int y, x;

	for ( y = 0; y < srcwin->lines && y < dstwin->lines; y++ )
		for ( x = 0; x < srcwin->cols && x < dstwin->cols; x++ )
			dstwin->winBuf[ y ][ x ] = srcwin->winBuf[ y ][ x ];
	return OK;
}

/* like a combination of overlay() and overwrite() that lets you     */
/* specify a region to copy, rather than just doing the whole window */

int copywin( WINDOW *src, WINDOW *dst, int sminrow, int smincol, int dminrow,
			  int dmincol, int dmaxrow, int dmaxcol, int overlay )
{
	chtype ch;
	int sy, sx, dy, dx;

	for ( sy = sminrow, dy = dminrow; sy < src->lines && dy < dmaxrow && dy < dst->lines;
		  sy++, dy++ )
	{
		for ( sx = smincol, dx = dmincol; sx < src->cols && dx < dmaxcol && dx < dst->cols;
			  sx++, dx++ )
		{
			ch = src->winBuf[ sy ][ sx ];
			if( overlay || ( ch & 0xff ) > ' ' )
				dst->winBuf[ dy ][ dx ] = ch;
		}
	}
	return OK;
}

int box( WINDOW *win, chtype vert, chtype horiz )
{
	int i;
	short wl, wc;
	chtype **buf;

	wl = win->lines - 1;
	wc = win->cols - 1;
	buf = win->winBuf;
	if ( vert == 0 && horiz == 0 )	/* get to do what we want */
	{
		for ( i = 1; i < wc; i++ )
		{	/* solid lines on top & bottom */
			buf[ 0 ][ i ] = buf[ wl ][ i ] = ACS_HLINE;
			/* buf[ 0 ][ i ] = buf[ wl ][ i ] = win->attrword | ACS_HLINE; */
		}
	
		for ( i = 1; i < wl; i++ )
		{
			/* solid lines on left & right */
			buf[ i ][ 0 ] = buf[ i ][ wc ] = ACS_VLINE;
			/* buf[ i ][ 0 ] = buf[ i ][ wc ] = win->attrword | ACS_VLINE; */
			win->begLine[ i ] = 0;
			win->endLine[ i ] = wc;
		}
		
		buf[ 0 ][ 0 ] = ACS_ULCORNER;
		buf[ wl ][ 0 ] = ACS_LLCORNER;
		buf[ 0 ][ wc ] = ACS_URCORNER;
		buf[ wl ][ wc ] = ACS_LRCORNER;
		/* buf[ 0 ][ 0 ] = win->attrword | ACS_ULCORNER;
		buf[ wl ][ 0 ] = win->attrword | ACS_LLCORNER;
		buf[ 0 ][ wc ] = win->attrword | ACS_URCORNER;
		buf[ wl ][ wc ] = win->attrword | ACS_LRCORNER; */
		win->begLine[ 0 ] = win->begLine[ wl ] = 0;
		win->endLine[ 0 ] = win->endLine[ wl ] = wc;
   }
   else
   {
		if ( vert == 0 )
			vert = ACS_VLINE;		/* boring box */
		if ( horiz == 0 )
	    	horiz = ACS_HLINE;

		for ( i = 1; i < wc; i++ )
			buf[ 0 ][ i ] = buf[ wl ][ i ] = horiz;
			/* buf[ 0 ][ i ] = buf[ wl ][ i ] = win->attrword | horiz; */
	
		for ( i = 0; i <= wl; i++ )
		{
			buf[ i ][ 0 ] = buf[ i ][ wc ] = vert;
			/* buf[ i ][ 0 ] = buf[ i ][ wc ] = win->attrword | vert; */
			win->begLine[ i ] = 0;
			win->endLine[ i ] = wc;
		}
	}
	return OK;
}

int scroll( WINDOW *win )
{
	if ( win == stdscr )
	{
		mac_scroll();
		scrollWindow( win, FALSE ); 	/* don't set update markers */
	}
	else
		scrollWindow( win, TRUE );
	return OK;
}

int scrollok( WINDOW *win, bool flag )
{
	win->f_scroll = flag;
	return OK;
}

int leaveok( WINDOW *win, bool flag )
{
	win->f_leave = flag;
	return OK;
}

int clearok( WINDOW *win, bool flag )
{
	win->f_clear = flag;
	return OK;
}

/* as you can see this does nothing */

int idlok( WINDOW *win, bool flag )
{
	return OK;
}

int nodelay( WINDOW *win, bool flag )
{
	win->f_nodelay = flag;
	return OK;
}

int cbreak( void )
{
	inputMode |= M_CBREAK;
	return OK;
}

int nocbreak( void )
{
	inputMode &= (~M_CBREAK);
	return OK;
}

int raw( void )
{
	inputMode |= M_RAW;
	return OK;
}

int noraw( void )
{
	inputMode &= (~M_RAW);
	return OK;
}

int echo( void )
{
	f_echo = TRUE;
	return OK;
}

int noecho( void )
{
	f_echo = FALSE;
	return OK;
}

int nl( void )
{
	f_nl = TRUE;
	return OK;
}

int nonl( void )
{
	f_nl = FALSE;
	return OK;
}

int baudrate( void )
{
	return 0;
}

char erasechar( void )
{
	return 0;
}

char killchar( void )
{
	return 0;
}

char *longname( void )
{
	return terminalName;
}

int wattrset( WINDOW *win, int attrs )
{
	win->attrword = attrs;
	return OK;
}

int beep( void )
{
	mac_beep();
	return OK;
}

int flash( void )
{
	return beep();
}

/* The only "keypad-style" characters I support now are the arrow keys */
/* and no special processing is needed since they return single byte   */
/* values in the range 0 - 0x1f */

int keypad( WINDOW *win, bool flag )
{
	return OK;
}

int wstandout( WINDOW *win )
{
	win->f_standout = TRUE;
	return OK;
}

int wstandend( WINDOW *win )
{
	win->f_standout = FALSE;
	return OK;
}

/* put blanks in every place in a window */

int werase( WINDOW *win )
{
	int x, y;
	short wl, wc;
	chtype **buf;

	wl = win->lines;
	wc = win->cols - 1;
	buf = win->winBuf;
	for ( y = 0; y < wl; y++ )
	{
		win->begLine[ y ] = 0;
		win->endLine[ y ] = wc;
		for( x = 0; x <= wc; x++ )
			buf[ y ][ x ] = ' ';
			/* buf[ y ][ x ] = win->attrword | ' '; */
	}
	win->curX = win->curY = 0;

	return OK;
}

/* werase() but also set f_clear, which makes wrefresh call the */
/* console driver clear screen and redraw the entire screen */

int wclear( WINDOW *win )
{
	win->f_clear = TRUE;
	return werase( win );
}

/* wdeleteln deletes the line the cursor is on and moves up */
/* the lines below it. */

int wdeleteln( WINDOW *win )
{
	int x, y;
	short wl, wc;
	chtype **buf;

	wl = win->lines;
	wc = win->cols - 1;
	buf = win->winBuf;

	win->begLine[ win->curY ] = 0;		
	win->endLine[ win->curY ] = wc;
	for ( y = win->curY + 1; y < wl; y++ )
	{
		win->begLine[ y ] = 0;		/* move all the lines after the current */
		win->endLine[ y ] = wc;		/* one back one line */
		for ( x = 0; x <= wc; x++ )
			buf[ y - 1 ][ x ] = buf[ y ][ x ];
	}

	for( x = 0; x < wc; x++ )
		buf[ wl - 1 ][ x ] = ' ';
		/* buf[ wl - 1 ][ x ] = win->attrword | ' '; */

	return OK;
}    

/* winsertln does the opposite. It makes a new blank line where the */
/* cursor is and pushed the current line and all subsequent lines   */
/* down. the bottom line is lost */

int winsertln( WINDOW *win )
{
	int x, y;
	short wl, wc, wcy;
	chtype **buf;

	wl = win->lines;
	wc = win->cols - 1;
	wcy = win->curY;
	buf = win->winBuf;

	for ( y = wl - 1; y > wcy; y-- )
	{
		win->begLine[ y ] = 0;
		win->endLine[ y ] = wc;
		for ( x = 0; x <= wc; x++ )
			buf[ y ][ x ] = buf[ y - 1 ][ x ];
	}

	win->begLine[ wcy ] = 0;
	win->endLine[ wcy ] = wc;
	for( x = 0; x <= wc; x++ )
		buf[ wcy ][ x ] = ' ';
		/* buf[ wcy ][ x ] = win->attrword | ' '; */

	return OK;
}    

/* insert a character at the cursor position and push the other stuff */
/* on the line to the right */

int winsch( WINDOW *win, chtype c )
{
	int x;
	short wcx, wcy;
	chtype **buf;

	wcy = win->curY;
	wcx = win->curX;
	buf = win->winBuf;

	for( x = win->cols - 1; x > wcx; x-- )		/* move current line right */
		buf[ wcy ][ x ] = buf[ wcy ][ x - 1 ];

	if ( wcx < win->begLine[ wcy ] )	/* set update markers for the line */
		win->begLine[ wcy ] = wcx;
		
	win->endLine[ wcy ] = win->cols - 1;
	consolePutchar( win, c );			/* put in the new character */

	return OK;
}

/* delete the character at the current cursor position and move */
/* other characters on the line to the left to take the space */

int wdelch( WINDOW *win )
{
	int x;
	short wcx, wcy;
	chtype **buf;

	wcy = win->curY;
	wcx = win->curX;
	buf = win->winBuf;

	for( x = wcx + 1; x < win->cols; x++ )
		buf[ wcy ][ x - 1 ] = buf[ wcy ][ x ];

	if ( wcx < win->begLine[ wcy ] )
		win->begLine[ wcy ] = wcx;
	win->endLine[ wcy ] = win->cols - 1;
	buf[ wcy ][ win->cols - 1 ] = ' ';
	/* buf[ wcy ][ win->cols - 1 ] = win->attrword | ' '; */

	return OK;
}

/* these are self-explanatory... */

int wclrtoeol( WINDOW *win )
{
	int x;
	short wcx, wcy;
	chtype **buf;

	wcy = win->curY;
	wcx = win->curX;
	buf = win->winBuf;

	for ( x = wcx; x < win->cols; x++ )
		buf[ wcy ][ x ] = ' ';
		/* buf[ wcy ][ x ] = win->attrword | ' '; */

	if ( wcx < win->begLine[ wcy ] )
		win->begLine[ wcy ] = wcx;
	win->endLine[ wcy ] = win->cols - 1;

	return OK;
}

int wclrtobot( WINDOW *win )
{
	int x, y;
	short wc, wcx, wcy;
	chtype **buf;

	wc = win->cols - 1;
	wcy = win->curY;
	wcx = win->curX;
	buf = win->winBuf;

	for ( x = wcx; x <= wc; x++ )
		buf[ wcy ][ x ] = ' ';
		/* buf[ wcy ][ x ] = win->attrword | ' '; */

	if ( wcx < win->begLine[ wcy ] )
		win->begLine[ wcy ] = wcx;
	win->endLine[ wcy ] = wc;

	for( y = wcy + 1; y < win->lines; y++ )
	{
		win->begLine[ y ] = 0;
		win->endLine[ y ] = wc;
		for( x = 0; x <= wc; x++ )
			buf[ y ][ x ] = ' ';
			/* buf[ y ][ x ] = win->attrword | ' '; */
	}

	return OK;
}

/* move the current cursor position for the window */

int wmove( WINDOW *win, int y, int x )
{
  if (y < 0) y = 0;
  else if (y >= win->lines) y = win->lines - 1;
  if (x < 0) x = 0;
  else if (x >= win->cols) x = win->cols - 1;

	win->curY = y;
	win->curX = x;
	return OK;
}

int waddch( WINDOW *win, chtype c )
{
	consolePutchar( win, c );
	return OK;
}

int waddstr( WINDOW *win, char *s )
{
	while ( *s != '\0' )
	{
		consolePutchar( win, (chtype)((unsigned char)(*s)) );
		s++;
	}

	return OK;
}

/* All the printw funcs use stdarg.h and call the stdio vsprintf() */
/* to print into a string buffer, which is then put out to the    */
/* console using vputchar() */

int printw( const char *fstr, ... )
{
	int count, i;
	va_list arglist;

	va_start( arglist, fstr );
	count = vsprintf( (char *)strBuf, fstr, arglist );
	va_end( arglist );

	for( i = 0; i < count; i++ )
		consolePutchar( stdscr, (chtype)(strBuf[ i ]) );

	return count;
}

int wprintw( WINDOW *win, const char *fstr, ... )
{
	int count, i;
	va_list arglist;

	va_start( arglist, fstr );
	count = vsprintf( (char *)strBuf, fstr, arglist );
	va_end( arglist );

	for( i = 0; i < count; i++ )
		consolePutchar( win, (chtype)(strBuf[ i ]) );

	return count;
}

int vwprintw( WINDOW *win, const char *fstr, va_list arg )
{
	int count, i;

	count = vsprintf( (char *)strBuf, fstr, arg );

	for( i = 0; i < count; i++ )
		consolePutchar( win, (chtype)(strBuf[ i ]) );

	return count;
}

int mvprintw( int y, int x, const char *fstr, ... )
{
	int count, i;
	va_list arglist;

	stdscr->curY = y;
	stdscr->curX = x;

	va_start( arglist, fstr );
	count = vsprintf( (char *)strBuf, fstr, arglist );
	va_end( arglist );

	for( i = 0; i < count; i++ )
		consolePutchar( stdscr, (chtype)(strBuf[ i ]) );

	return count;
}

int mvwprintw( WINDOW *win, int y, int x, const char *fstr, ... )
{
	int count, i;
	va_list arglist;

	win->curY = y;
	win->curX = x;

	va_start( arglist, fstr );
	count = vsprintf( (char *)strBuf, fstr, arglist );
	va_end( arglist );

	for( i = 0; i < count; i++ )
		consolePutchar( win, (chtype)(strBuf[ i ]) );

	return count;
}

chtype wgetch( WINDOW *win )
{
	return (chtype)consoleGetchar( win );
}

char *wgetstr( WINDOW *win, char *s )
{
	int ch;
	char *p;

	p = s;
	ch = consoleGetchar( win );
	while( ch != '\0' && ch != LF )
	{
		*p++ = (char)ch;
		ch = consoleGetchar( win );
	}
	*p = '\0';
	return p;
}

/* wnoutrefresh and doupdate are here mainly for some kind of completeness */
/* using them is actually slower than just calling wrefresh, because my    */
/* wrefresh is based on the "do less computer work" model, rather than the */
/* "send fewer characters to the terminal" model. */

int wnoutrefresh( WINDOW *win )
{
	short wt, wl;
	short *wbl, *wel;
	int x, y;
	chtype **sbuf, **wbuf;

	wt = win->top;
	wl = win->left;
	wbl = win->begLine;
	wel = win->endLine;
	sbuf = stdscr->winBuf;
	wbuf = win->winBuf;

	if ( win->f_clear )
	{
		werase( curscr );
		mac_clear();
		win->f_clear = FALSE;
	}

	for ( y = 0; y < win->lines; y++ )
	{
		if ( wel[ y ] > 0 )
		{
			for ( x = wbl[ y ]; x <= wel[ y ]; x++ )
				sbuf[ wt + y ][ wl + x ] = wbuf[ y ][ x ];

			wel[ y ] = -1;
			wbl[ y ] = win->cols;
		}
	}
	return OK;
}

int doupdate( void )
{
	chtype c;
	int x, y, charcount;
	chtype **sbuf, **cbuf;

	sbuf = stdscr->winBuf;
	cbuf = curscr->winBuf;

	charcount = 0;
	for ( y = 0; y < LINES; y++ )
	{
		for ( x = 0; x < COLS; x++ )
		{
			c = sbuf[ y ][ x ];
			if ( c != cbuf[ y ][ x ] )
			{
				cbuf[ y ][ x ] = c;
				mac_putchar( c, x, y );
				mac_update( y, x, x );
				charcount++;
			}
		}
	}
	return charcount;
}

/*
This would be the standard curses wrefresh() but it's slower than mine
below. Curses was designed with the idea that the computer was fast
and the terminal connection was slow: 4800 baud or some such, so it
is designed to minimize output to the terminal. This actually makes
it slower than it has to be when output to the terminal is as fast
as the computer.

int wrefresh( WINDOW *win )
{
	wnoutrefresh( win );
	return doupdate();
}
*/

int wrefresh( WINDOW *win )
{
	short wt, wl;
	short *wbl, *wel;
	short x, y;
	chtype c;
	int charcount;
	chtype **cbuf, **wbuf;

	wt = win->top;			/* do ptr derefs once, instead of a zillion times*/
	wl = win->left;
	wbl = win->begLine;
	wel = win->endLine;
	cbuf = curscr->winBuf;
	wbuf = win->winBuf;

	charcount = 0;
	if ( win->f_clear )		/* check for f_clear flag. If it's set	*/
	{						/* wipe curscr to force a full screen	*/
		werase( curscr );	/* redraw and then clear screen    */
		mac_clear();
		win->f_clear = FALSE;
	}

	for ( y = 0; y < win->lines; y++ )
	{
		if ( wel[ y ] > 0 )		/* go through the update arrays and draw */
		{							/* the parts of the lines marked */
			for ( x = wbl[ y ]; x <= wel[ y ]; x++ )
			{
				c = wbuf[ y ][ x ];		/* output the characters */
				mac_putchar( c, wl + x, wt + y );
				charcount++;
				cbuf[ wt + y ][ wl + x ] = c;
			}
			mac_update( wt + y, wl + wbl[ y ], wl + wel[ y ] );
			wel[ y ] = -1;				/* no more update needed on this line */
			wbl[ y ] = win->cols;		/* for now... */
		}
	}

	return charcount;
}
